package ru.dbotthepony.mc.otm.capability;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.core.SectionPos;
import net.minecraft.server.level.ServerLevel;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.entity.BlockEntity;
import net.minecraft.world.level.chunk.LevelChunk;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.storage.IStorageIdentity;
import ru.dbotthepony.mc.otm.storage.StorageGrid;

import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.util.List;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public interface IStorageGridCell {
	/**
	 * Return all providers that this cell can store. Or nothing, if it is not a storage.
	 *
	 * If you don't know what you are going to store, return empty list, then, when you make up your
	 * mind by either player tweaking settings or something,
	 * call helper method detachStorage();
	 * add whatever going to end up in getComponents();
	 * add helper method attachStorage();
	 *
	 * Failing to do it in this order will cause IllegalStateException
	 *
	 * @return a list of all providers, inheriting IStorageIdentity, or an empty list
	 */
	List<? extends IStorageIdentity<?>> getComponents();

	default List<? extends IStorageIdentity<?>> getComponentsForView() {
		return getComponents();
	}

	default List<? extends IStorageIdentity<?>> getComponentsForInterfaces() {
		return getComponents();
	}

	default List<? extends IStorageIdentity<?>> getComponentsForExporters() {
		return getComponents();
	}

	default void detachStorage() {
		var grid = getStorageGrid();

		if (grid != null) {
			grid.detach(this);
		}
	}

	default void attachStorage() {
		var grid = getStorageGrid();

		if (grid != null) {
			grid.attach(this);
		}
	}

	@Nullable
	StorageGrid getStorageGrid();
	void setStorageGrid(@Nullable StorageGrid grid);
	boolean isValidStorageCell();

	default void scheduleDiscoverNeighbours(Level level, BlockPos pos) {
		if (level instanceof ServerLevel)
			StorageGrid.scheduleDiscoverNeighbours(this, level, pos);
	}

	default boolean connectOrCreateStorageGrid(Level level, BlockPos pos) {
		return connectOrCreateStorageGrid(level, pos, false);
	}

	default void onNeighbourStorageCell(Level level, BlockPos pos, Direction direction, IStorageGridCell cell) {

	}

	default boolean connectOrCreateStorageGrid(Level level, BlockPos pos, boolean force) {
		OverdriveThatMatters.LOGGER.info("Discover {}", this);
		if (getStorageGrid() != null && !force)
			return true;

		boolean full_discovery = true;

		for (Direction direction : Direction.values()) {
			BlockPos offset = pos.offset(direction.getNormal());

			// level.getBlockEntity can drink big cup of deadlocks
			LevelChunk get_chunk = level.getChunkSource().getChunkNow(SectionPos.blockToSectionCoord(offset.getX()), SectionPos.blockToSectionCoord(offset.getZ()));

			if (get_chunk == null) {
				full_discovery = false;
				continue;
			}

			BlockEntity get_entity = get_chunk.getBlockEntity(offset);

			if (get_entity != null && get_entity.getCapability(MatteryCapability.STORAGE_CELL, direction.getOpposite()).isPresent()) {
				var cell = get_entity.getCapability(MatteryCapability.STORAGE_CELL, direction.getOpposite()).resolve().get();
				var grid = cell.getStorageGrid();

				if (grid != null && grid != getStorageGrid()) {
					if (getStorageGrid() == null) {
						grid.add(this);
					} else {
						grid.merge(getStorageGrid());
					}
				}

				if (cell.isValidStorageCell())
					onNeighbourStorageCell(level, pos, direction, cell);
			}
		}

		if (getStorageGrid() == null) {
			new StorageGrid().add(this);
		}

		return full_discovery;
	}
}
